from typing import Any

from .._dto import VTEXDataResponse, VTEXItemsResponse
from .._types import DictType
from .base import BaseAPI


class PromotionsAndTaxesAPI(BaseAPI):
    """
    Client for the Promotions and Taxes API.
    https://developers.vtex.com/docs/api-reference/promotions-and-taxes-api
    """

    ENVIRONMENT = "vtexcommercestable"

    def list_archived_promotions(
        self,
        **kwargs: Any,
    ) -> VTEXItemsResponse[DictType, DictType]:
        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint="api/rnb/pvt/archive/benefits/calculatorconfiguration",
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXItemsResponse[DictType, DictType],
        )

    def get_promotions(self, **kwargs: Any) -> VTEXDataResponse[DictType]:
        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint="api/rnb/pvt/benefits/calculatorconfiguration",
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[DictType],
        )

    def get_taxes(self, **kwargs: Any) -> VTEXDataResponse[DictType]:
        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint="api/rnb/pvt/taxes/calculatorconfiguration",
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[DictType],
        )

    def get_promotion_or_tax(
        self,
        promotion_or_tax_id: str,
        **kwargs: Any,
    ) -> VTEXDataResponse[DictType]:
        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint=f"/api/rnb/pvt/calculatorconfiguration/{promotion_or_tax_id}",
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[DictType],
        )



from datetime import datetime, timezone, tzinfo
from re import compile
from typing import Any, Dict, Mapping, Union
from uuid import UUID

from dateutil.parser import parse as parse_datetime
from dateutil.tz import tzoffset
from distutils.util import strtobool

from ._constants import APP_KEY_HEADER, APP_TOKEN_HEADER
from ._logging import UTILS_LOGGER
from ._sentinels import UNDEFINED

TO_SNAKE_CASE_STEP_1_PATTERN = compile(r"(.)([A-Z][a-z]+)")
TO_SNAKE_CASE_STEP_2_PATTERN = compile(r"([a-z0-9])([A-Z])")


def is_nullish_str(value: str) -> bool:
    return value.lower() in {"", "null", "none", "nil"}


def exclude_undefined_values(obj: Dict[Any, Any]) -> Dict[Any, Any]:
    return {key: value for key, value in obj.items() if value is not UNDEFINED}


def str_to_bool(value: str) -> bool:
    return bool(strtobool(value))


def remove_null_bytes(value: str) -> str:
    return value.replace("\x00", "")


def to_snake_case(string: str) -> str:
    try:
        UUID(string)
    except (AttributeError, ValueError, TypeError):
        return TO_SNAKE_CASE_STEP_2_PATTERN.sub(
            r"\1_\2",
            TO_SNAKE_CASE_STEP_1_PATTERN.sub(r"\1_\2", string),
        ).lower()

    return string


def to_snake_case_deep(obj: Any) -> Any:
    if isinstance(obj, dict):
        snake_cased_obj = {}

        for key, value in sorted(obj.items(), key=lambda item: item[0]):
            if isinstance(key, str):
                key = to_snake_case(remove_null_bytes(key))

                while key in snake_cased_obj:
                    UTILS_LOGGER.debug(
                        f"Snake cased key {key} appears multiple times in object",
                        extra={"object": obj, "key": key},
                    )
                    key = f"_{key}"

            snake_cased_obj[key] = to_snake_case_deep(value)

        return snake_cased_obj

    if isinstance(obj, (list, set, tuple)):
        return type(obj)([to_snake_case_deep(element) for element in obj])

    if isinstance(obj, str):
        return remove_null_bytes(obj)

    return obj


def redact_headers(headers: Mapping[str, str]) -> Dict[str, str]:
    redacted_headers = {}

    for key, value in list(headers.items()):
        if key.lower() in {APP_KEY_HEADER.lower(), APP_TOKEN_HEADER.lower()}:
            redacted_headers[key] = "*" * 32
        else:
            redacted_headers[key] = value

    return redacted_headers


def to_tzinfo(tz: Union[tzinfo, int, None] = None) -> tzinfo:
    if isinstance(tz, tzinfo):
        return tz

    if isinstance(tz, int):
        return tzoffset(None, tz)

    return timezone.utc


def to_datetime(
    value: Union[datetime, str],
    use_tz: bool = True,
    tz: Union[tzinfo, int, None] = None,
) -> datetime:
    value_as_datetime = value if isinstance(value, datetime) else parse_datetime(value)

    if not use_tz:
        return value_as_datetime.replace(tzinfo=None)

    if value_as_datetime.tzinfo and tz is None:
        return value_as_datetime

    return value_as_datetime.replace(tzinfo=to_tzinfo(tz))


def now(use_tz: bool = True, tz: Union[tzinfo, None] = None) -> datetime:
    return datetime.now(to_tzinfo(tz) if use_tz else None)


def three_years_ago(use_tz: bool = True, tz: Union[tzinfo, None] = None) -> datetime:
    current_datetime = now(use_tz=use_tz, tz=tz)

    return current_datetime.replace(
        year=current_datetime.year - 3,
        hour=0,
        minute=0,
        second=0,
        microsecond=0,
    )



from functools import cached_property
from typing import TYPE_CHECKING, List, Union

from ._config import Config  # type: ignore[attr-defined]
from ._logging import CLIENT_LOGGER
from ._sentinels import UNDEFINED, UndefinedSentinel

if TYPE_CHECKING:
    from ._api import (
        CatalogAPI,
        CheckoutAPI,
        CustomAPI,
        LicenseManagerAPI,
        LogisticsAPI,
        MasterDataAPI,
        OrdersAPI,
        PaymentsGatewayAPI,
        PromotionsAndTaxesAPI,
    )


class VTEX:
    """
    Entrypoint for the VTEX SDK.
    From this class you can access all the APIs on VTEX
    """

    def __init__(
        self,
        account_name: Union[str, UndefinedSentinel] = UNDEFINED,
        app_key: Union[str, UndefinedSentinel] = UNDEFINED,
        app_token: Union[str, UndefinedSentinel] = UNDEFINED,
        timeout: Union[float, None, UndefinedSentinel] = UNDEFINED,
        retry_attempts: Union[int, UndefinedSentinel] = UNDEFINED,
        retry_backoff_min: Union[float, UndefinedSentinel] = UNDEFINED,
        retry_backoff_max: Union[float, UndefinedSentinel] = UNDEFINED,
        retry_backoff_exponential: Union[bool, float, UndefinedSentinel] = UNDEFINED,
        retry_statuses: Union[List[int], UndefinedSentinel] = UNDEFINED,
        retry_logs: Union[bool, UndefinedSentinel] = UNDEFINED,
        raise_for_status: Union[bool, UndefinedSentinel] = UNDEFINED,
    ) -> None:
        self.logger = CLIENT_LOGGER
        self.config = Config(
            account_name=account_name,
            app_key=app_key,
            app_token=app_token,
            timeout=timeout,
            retry_attempts=retry_attempts,
            retry_backoff_min=retry_backoff_min,
            retry_backoff_max=retry_backoff_max,
            retry_backoff_exponential=retry_backoff_exponential,
            retry_statuses=retry_statuses,
            retry_logs=retry_logs,
            raise_for_status=raise_for_status,
        )

    @cached_property
    def custom(self) -> "CustomAPI":
        from ._api import CustomAPI

        return CustomAPI(client=self)

    @cached_property
    def catalog(self) -> "CatalogAPI":
        from ._api import CatalogAPI

        return CatalogAPI(client=self)

    @cached_property
    def checkout(self) -> "CheckoutAPI":
        from ._api import CheckoutAPI

        return CheckoutAPI(client=self)

    @cached_property
    def license_manager(self) -> "LicenseManagerAPI":
        from ._api import LicenseManagerAPI

        return LicenseManagerAPI(client=self)

    @cached_property
    def logistics(self) -> "LogisticsAPI":
        from ._api import LogisticsAPI

        return LogisticsAPI(client=self)

    @cached_property
    def master_data(self) -> "MasterDataAPI":
        from ._api import MasterDataAPI

        return MasterDataAPI(client=self)

    @cached_property
    def orders(self) -> "OrdersAPI":
        from ._api import OrdersAPI

        return OrdersAPI(client=self)

    @cached_property
    def payments_gateway(self) -> "PaymentsGatewayAPI":
        from ._api import PaymentsGatewayAPI

        return PaymentsGatewayAPI(client=self)

    @cached_property
    def promotions_and_taxes(self) -> "PromotionsAndTaxesAPI":
        from ._api import PromotionsAndTaxesAPI

        return PromotionsAndTaxesAPI(client=self)



from dataclasses import dataclass
from json import JSONDecodeError
from math import ceil
from typing import Any, Dict, Generic, List, TypeVar, Union

from httpx import Request, Response

from ._utils import remove_null_bytes, to_snake_case_deep

VTEXResponseType = TypeVar("VTEXResponseType", bound="VTEXResponse")
DataType = TypeVar("DataType", bound=Any)
ItemType = TypeVar("ItemType", bound=Any)


@dataclass
class VTEXRequest:
    request: Request
    method: str
    url: str
    headers: Dict[str, str]

    @classmethod
    def factory(cls, request: Request) -> "VTEXRequest":
        return cls(
            request=request,
            method=str(request.method).upper(),
            url=str(request.url),
            headers=dict(request.headers),
        )


@dataclass
class VTEXResponse:
    request: VTEXRequest
    response: Response
    status: int
    headers: Dict[str, str]

    @classmethod
    def factory(cls, response: Response) -> "VTEXResponse":
        return cls(
            request=VTEXRequest.factory(response.request),
            response=response,
            status=int(response.status_code),
            headers=dict(response.headers),
        )


@dataclass
class VTEXDataResponse(VTEXResponse, Generic[DataType]):
    data: DataType

    @classmethod
    def get_original_response_data(cls, response: Response) -> Any:
        try:
            return to_snake_case_deep(response.json(strict=False))
        except JSONDecodeError:
            return remove_null_bytes(response.text)

    @classmethod
    def factory(cls, response: Response) -> "VTEXDataResponse[DataType]":
        vtex_response = VTEXResponse.factory(response)

        return cls(
            request=vtex_response.request,
            response=vtex_response.response,
            status=vtex_response.status,
            headers=vtex_response.headers,
            data=cls.get_original_response_data(response),
        )


@dataclass
class VTEXItemsResponse(VTEXDataResponse[DataType], Generic[DataType, ItemType]):
    items: List[ItemType]

    @classmethod
    def factory(cls, response: Response) -> "VTEXItemsResponse[DataType, ItemType]":
        vtex_data_response = VTEXDataResponse[DataType].factory(response)
        data = vtex_data_response.data

        if isinstance(data, list):
            items = data
        elif isinstance(data, dict) and isinstance(data.get("list"), list):
            items = data["list"]
        elif isinstance(data, dict) and isinstance(data.get("items"), list):
            items = data["items"]
        elif isinstance(data, dict) and isinstance(data.get("data"), list):
            items = data["data"]
        else:
            raise ValueError(f"Not a valid items response: {data}")

        return cls(
            request=vtex_data_response.request,
            response=vtex_data_response.response,
            status=vtex_data_response.status,
            headers=vtex_data_response.headers,
            data=vtex_data_response.data,
            items=list(items),
        )


@dataclass
class VTEXPagination:
    total: int
    pages: int
    page_size: int
    page: int
    previous_page: Union[int, None]
    next_page: Union[int, None]

    @classmethod
    def factory(
        cls,
        vtex_items_response: VTEXItemsResponse[DataType, ItemType],
    ) -> "VTEXPagination":
        data = vtex_items_response.data
        request_headers = vtex_items_response.request.headers
        response_headers = vtex_items_response.headers

        total, pages, page_size, page = -1, -1, -1, -1
        if isinstance(data, dict) and data.get("paging"):
            pagination = data["paging"]
            total = pagination["total"]
            page_size = pagination["per_page"]
            pages = pagination["pages"]
            page = int(pagination.get("page") or pagination.get("current_page"))
        elif isinstance(data, dict) and data.get("total_page"):
            total = data["total_rows"]
            page_size = data["size"]
            pages = data["total_page"]
            page = data["page"]
        elif "rest-content-range" in response_headers:
            request_pagination = request_headers["rest-range"].split("=")[-1].split("-")
            response_pagination = response_headers["rest-content-range"].split(" ")[-1]
            total = int(response_pagination.split("/")[1])
            page_size = int(request_pagination[1]) - int(request_pagination[0])
            pages = ceil(total / page_size)
            page = ceil(
                int(response_pagination.split("/")[0].split("-")[1]) / page_size
            )

        if all(
            isinstance(field, int) and field != -1
            for field in {total, pages, page_size, page}
        ):
            return cls(
                total=total,
                pages=pages,
                page_size=page_size,
                page=page,
                previous_page=page - 1 if page > 1 else None,
                next_page=page + 1 if page < pages else None,
            )

        raise ValueError(f"Not a valid paginated items response: {vtex_items_response}")


@dataclass
class VTEXPaginatedItemsResponse(VTEXItemsResponse[DataType, ItemType]):
    pagination: VTEXPagination

    @classmethod
    def factory(cls, response: Response) -> "VTEXItemsResponse[DataType, ItemType]":
        vtex_items_response = VTEXItemsResponse[DataType, ItemType].factory(response)

        return cls(
            request=vtex_items_response.request,
            response=vtex_items_response.response,
            data=vtex_items_response.data,
            status=vtex_items_response.status,
            headers=vtex_items_response.headers,
            items=vtex_items_response.items,
            pagination=VTEXPagination.factory(vtex_items_response),
        )


@dataclass
class VTEXCartItem:
    sku_id: str
    quantity: int
    seller_id: str

    def to_vtex_cart_item(self) -> Dict[str, Union[str, int]]:
        return {
            "id": self.sku_id,
            "quantity": self.quantity,
            "seller": self.seller_id,
        }



from typing import Any, List, Union

from .._dto import VTEXCartItem, VTEXDataResponse
from .._sentinels import UNDEFINED, UndefinedSentinel
from .._types import DictType
from .._utils import exclude_undefined_values
from .base import BaseAPI


class CheckoutAPI(BaseAPI):
    """
    Client for the Catalog API.
    https://developers.vtex.com/docs/api-reference/checkout-api
    """

    ENVIRONMENT = "vtexcommercestable"

    def run_cart_simulation(
        self,
        cart: List[VTEXCartItem],
        country: Union[str, UndefinedSentinel] = UNDEFINED,
        postal_code: Union[str, UndefinedSentinel] = UNDEFINED,
        geo_coordinates: Union[List[float], UndefinedSentinel] = UNDEFINED,
        rnb_behavior: Union[int, UndefinedSentinel] = UNDEFINED,
        sales_channel: Union[int, UndefinedSentinel] = UNDEFINED,
        individual_shipping_estimates: Union[bool, UndefinedSentinel] = UNDEFINED,
        **kwargs: Any,
    ) -> VTEXDataResponse[DictType]:
        return self._request(
            method="POST",
            environment=self.ENVIRONMENT,
            endpoint="/api/checkout/pub/orderForms/simulation",
            params=exclude_undefined_values({
                "RnbBehavior": rnb_behavior,
                "sc": sales_channel,
                "individualShippingEstimates": individual_shipping_estimates,
            }),
            json=exclude_undefined_values({
                "items": [item.to_vtex_cart_item() for item in cart],
                "country": country,
                "postalCode": postal_code,
                "geoCoordinates": geo_coordinates,
            }),
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[DictType],
        )



from typing import Any, List, Union

from .._constants import LIST_ROLES_MAX_PAGE_SIZE, LIST_ROLES_START_PAGE, MIN_PAGE_SIZE
from .._dto import VTEXDataResponse, VTEXItemsResponse, VTEXPaginatedItemsResponse
from .._sentinels import UNDEFINED, UndefinedSentinel
from .._types import OrderingDirectionType
from .base import BaseAPI
from .types.license_manager import GetAccountData, ListRolesData, Role, UserRole


class LicenseManagerAPI(BaseAPI):
    """
    Client for the License Manager API.
    https://developers.vtex.com/docs/api-reference/license-manager-api
    """

    ENVIRONMENT = "vtexcommercestable"

    def get_account(self, **kwargs: Any) -> VTEXDataResponse[GetAccountData]:
        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint="/api/vlm/account",
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[GetAccountData],
        )

    def get_user_roles(
        self,
        user_id: Union[str, UndefinedSentinel] = UNDEFINED,
        **kwargs: Any,
    ) -> VTEXItemsResponse[List[UserRole], UserRole]:
        config = self.client.config.with_overrides(**kwargs)

        if not user_id:
            app_keys = {
                app_key["app_key"]: app_key
                for app_key in self.get_account().data["app_keys"]
            }

            user_id = app_keys[config.get_app_key()]["id"]

        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint=f"/api/license-manager/users/{user_id}/roles",
            config=config,
            response_class=VTEXItemsResponse[List[UserRole], UserRole],
        )

    def list_roles(
        self,
        order_by_field: str = "id",
        order_by_direction: OrderingDirectionType = "DESC",
        page: int = LIST_ROLES_START_PAGE,
        page_size: int = LIST_ROLES_MAX_PAGE_SIZE,
        **kwargs: Any,
    ) -> VTEXPaginatedItemsResponse[ListRolesData, Role]:
        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint="/api/license-manager/site/pvt/roles/list/paged",
            params={
                "sort": order_by_field,
                "sortType": order_by_direction,
                "pageNumber": max(page, LIST_ROLES_START_PAGE),
                "numItems": max(
                    min(page_size, LIST_ROLES_MAX_PAGE_SIZE),
                    MIN_PAGE_SIZE,
                ),
            },
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXPaginatedItemsResponse[ListRolesData, Role],
        )



from typing import TypedDict


class CurrentPagePagination(TypedDict, total=True):
    current_page: int
    pages: int
    per_page: int
    total: int


class PagePagination(TypedDict, total=True):
    page: int
    pages: int
    per_page: int
    total: int


class RowsPagination(TypedDict, total=True):
    page: int
    size: int
    total_page: int
    total_rows: int



from typing import Any, Dict, Union

from httpx import Headers


class VTEXError(Exception):
    def __init__(self, *args: Any, **kwags: Any) -> None:
        super().__init__(*args, **kwags)


class VTEXRequestError(VTEXError):
    def __init__(
        self,
        *args: Any,
        exception: Union[Exception, None] = None,
        method: Union[str, None] = None,
        url: Union[str, None] = None,
        headers: Union[Headers, Dict[str, str], None] = None,
        **kwargs: Any,
    ) -> None:
        self.args = args
        self.exception = exception
        self.method = method
        self.url = url
        self.headers = dict(headers) if headers else None
        self.kwargs = kwargs

        super().__init__(str(exception or "VTEXRequestError"), *args, **kwargs)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "exception": self.exception,
            "method": self.method,
            "url": self.url,
            "headers": self.headers,
        }


class VTEXResponseError(VTEXError):
    def __init__(
        self,
        *args: Any,
        method: Union[str, None] = None,
        url: Union[str, None] = None,
        request_headers: Union[Headers, Dict[str, str], None] = None,
        status: Union[int, None] = None,
        data: Any = None,
        response_headers: Union[Headers, Dict[str, str], None] = None,
        **kwargs: Any,
    ) -> None:
        self.args = args
        self.method = method
        self.url = url
        self.request_headers = dict(request_headers) if request_headers else None
        self.status = status
        self.data = data
        self.response_headers = dict(response_headers) if response_headers else None
        self.kwargs = kwargs

        super().__init__(str(data or "VTEXResponseError"), *args, **kwargs)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "method": self.method,
            "url": self.url,
            "request_headers": self.request_headers,
            "status": self.status,
            "data": self.data,
            "response_headers": self.response_headers,
        }
